# SPDX-FileCopyrightText: © 2018 Bastien Pietropaoli
# SPDX-FileCopyrightText: © 2025 Evotis S.A.S.
# SPDX-License-Identifier: Apache-2.0

"""
Copyright (c) 2018 Bastien Pietropaoli

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

All additions and modifications are Copyright (c) 2025 Evotis S.A.S.
"""

import json
import logging
import re
import warnings
from typing import Any, Callable, ClassVar, Dict, Type, TypeVar, cast

from typing_extensions import override

import __main__
from kajson.exceptions import UnijsonEncoderError
from kajson.sandbox_manager import sandbox_manager

ENCODER_LOGGER_CHANNEL_NAME = "kajson.encoder"
ENCODER_LOGGER_CHANNEL_NAME_IN_SANDBOX = "kajson.encoder.sandbox"
IS_ENCODER_FALLBACK_ENABLED = False
FALLBACK_MESSAGE = " Trying something else."


T = TypeVar("T")


class UniversalJSONEncoder(json.JSONEncoder):
    """
    A universal JSON encoder for Python objects. This encoder will work with
    simple custom objects by default and can automatically take into account methods
    `__json_encoder__()` defined in custom classes if implemented correctly. Those
    methods should return a dictionnary (with only strings as keys) and take only
    `self` as an argument.

    In addition, it is possible to register functions for types over which you have
    no control (standard / external library types). For this, use static method
    `UniversalJSONEncoder.register()`.

    The encoder adds attributes `__module__` and `__class__` to the resulting JSON
    objects to enable the identification of the object from which they were constructed.
    They both can be added directly into your encoder to use the same encoder for
    multiple objects (by setting it to a superclass for instance). See the encoders
    for timezones for an example of this. If your encoder already sets them, then they
    won't be modified in this universal encoder. Your values will stay.

    How to use this class:
        `json.dumps(obj, cls=UniversalJSONEncoder)`
                        OR
        `kajson.dumps(obj)`
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(ENCODER_LOGGER_CHANNEL_NAME)

    def log(self, message: str) -> None:
        if sandbox_manager.is_in_sandbox():
            logger = logging.getLogger(ENCODER_LOGGER_CHANNEL_NAME_IN_SANDBOX)
            logger.debug(message)
        else:
            self.logger.debug(message)

    # The registered encoding functions:
    _encoders: ClassVar[Dict[Type[Any], Callable[[Any], Dict[str, Any]]]] = {}

    @staticmethod
    def register(obj_type: Type[T], encoding_function: Callable[[T], Dict[str, Any]]) -> None:
        """
        Register a function as an encoder for the provided type/class. The provided
        encoder should take a single argument (the object to serialise) and return
        a JSON serialisable dictionnary (by the standard of this serialiser).
        Passing a new encoding function to a type already registered will overwrite
        the previously registered encoding function.
        Args:
            type (obj_type): The type to be encoded by the provided encoder. Can be
                easily obtained by simply providing a class directly.
            encoding_function (function): The function to use as an encoder for the
                provided type. Takes a single argument, a returns a dictionnary.
        """
        if not isinstance(obj_type, type):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise ValueError("Expected a type/class, a %s was passed instead." % type(obj_type))
        if not callable(encoding_function):
            raise ValueError("Expected a function, a %s was passed instead." % type(encoding_function))

        UniversalJSONEncoder._encoders[obj_type] = encoding_function

    # argument must be named "o" to override the default method
    @override
    def default(self, o: Any) -> Dict[str, Any]:
        """
        Extends the default behaviour of the default JSON encoder. It will try
        the different methods to encode the provided object in the following order:
         - Default JSON encoder (for known types)
         - Registered encoding function (if one is found)
         - `__json_encode__()` as provided by the custom class (if it's found)
         - Use the default __dict__ property of the object (for custom classes)
        Args:
            o (object): The object to serialise.
        Return:
            dict - A dictionnary of JSON serialisable objects.
        Raises:
            TypeError - If none of the methods worked.
        """
        obj: Any = o
        # Default JSON encoder:
        try:
            return cast(Dict[str, Any], json.JSONEncoder.default(self, obj))
        except TypeError:
            pass

        already_encoded = False
        the_dict: Dict[str, Any] = {}

        # Use a registered encoding function:
        if type(obj) in UniversalJSONEncoder._encoders:
            try:
                the_dict = UniversalJSONEncoder._encoders[type(obj)](obj)
                already_encoded = True
            except Exception as exc:
                func_name = UniversalJSONEncoder._encoders[type(obj)].__name__
                error_msg = f"Encoding function {func_name} used for type '{type(obj)}' raised an exception: {exc}."
                if IS_ENCODER_FALLBACK_ENABLED:
                    warnings.warn(error_msg + FALLBACK_MESSAGE)
                else:
                    raise UnijsonEncoderError(error_msg) from exc

        # Trying to use __json_encode__():
        if not already_encoded:
            try:
                the_dict = obj.__json_encode__()  # type: ignore
                already_encoded = True
            except AttributeError:  # No method __json_encode__() found
                pass
            except Exception as exc:
                error_msg = f"Method __json_encode__() used for type '{type(obj)}' raised an exception."
                if IS_ENCODER_FALLBACK_ENABLED:
                    warnings.warn(error_msg + FALLBACK_MESSAGE)
                else:
                    raise UnijsonEncoderError(error_msg) from exc

        # Trying the default __dict__ attribute:
        if not already_encoded:
            try:
                the_dict = dict(obj.__dict__)  # type: ignore
                already_encoded = True
            except AttributeError:
                pass

        # If nothing worked, raise an exception like the default JSON encoder would:
        if not already_encoded:
            raise TypeError(f"Type {type(obj)} is not JSON serializable. Value: {obj}")

        # Add the metadata used to reconstruct the object (if necessary):
        if "__class__" not in the_dict:
            the_dict["__class__"] = str(obj.__class__.__name__)
        if "__module__" not in the_dict:
            the_dict["__module__"] = _get_object_module(obj)

        return the_dict


#########################################################################################
#########################################################################################
#########################################################################################


def _get_object_module(obj: Any) -> str:
    """
    Get the name of the module from which the given object was created.
    Args:
        obj (object): Any object.
    Return:
        str - The name of the module from which the given object was created.
    """
    try:
        module_name = str(obj.__module__)
    except AttributeError:
        module_name = _get_type_module(obj.__class__)
    if module_name == "__main__":
        module_name = __main__.__file__[:-3]  # Prevents having __main__ as a module.
    return module_name

    # Remark 1: Builtin objects don't have a __module__ attribute.
    # Remark 2: inspect.getmodule(obj) should work but it doesn't.


def _get_type_module(the_type: Type[Any]) -> str:
    """
    Get the name of the module containing the given type.
    Args:
        t (type): The type for which the module is requested.
    Return:
        str - The name of the module containing the given type.
    """
    # 1) Extract the name of the module from str(type).
    # 2) Get the chain of submodules separated by dots.
    # 3) Join them together while getting rid of the last one.
    # Expressions used to find module names:
    __class_expression = re.compile(r"^<class '([a-zA-Z0-9._]*)'>")
    __type_expression = re.compile(r"^<type '([a-zA-Z0-9._]*)'>")
    the_type_str = str(the_type)
    if search_result := __class_expression.search(the_type_str):
        return ".".join(search_result.group(1).split(".")[:-1])
    elif search_result := __type_expression.search(the_type_str):
        return ".".join(search_result.group(1).split(".")[:-1])
    else:
        return "builtins"



# SPDX-FileCopyrightText: © 2025 Evotis S.A.S.
# SPDX-License-Identifier: Apache-2.0

import logging
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional, Type

from pydantic import BaseModel, Field, PrivateAttr, RootModel

from kajson.exceptions import ClassRegistryInheritanceError, ClassRegistryNotFoundError
from kajson.module_utils import find_classes_in_module, import_module_from_file
from kajson.sandbox_manager import sandbox_manager

LOGGING_LEVEL_VERBOSE = 5
CLASS_REGISTRY_LOGGER_CHANNEL_NAME = "class_registry"
CLASS_REGISTRY_LOGGER_CHANNEL_NAME_IN_SANDBOX = "class_registry.sandbox"

ClassRegistryDict = Dict[str, Type[Any]]


def find_files_in_dir(dir_path: str, pattern: str, is_recursive: bool) -> List[Path]:
    """
    Find files matching a pattern in a directory.

    Args:
        dir_path: Directory path to search in
        pattern: File pattern to match (e.g. "*.py")
        recursive: Whether to search recursively in subdirectories

    Returns:
        List of matching Path objects
    """
    path = Path(dir_path)
    if is_recursive:
        return list(path.rglob(pattern))
    else:
        return list(path.glob(pattern))


class ClassRegistry(RootModel[ClassRegistryDict]):
    root: ClassRegistryDict = Field(default_factory=dict)
    _logger: logging.Logger = PrivateAttr(logging.getLogger(CLASS_REGISTRY_LOGGER_CHANNEL_NAME))

    def log(self, message: str) -> None:
        if sandbox_manager.is_in_sandbox():
            logger = logging.getLogger(CLASS_REGISTRY_LOGGER_CHANNEL_NAME_IN_SANDBOX)
            logger.debug(message)
        else:
            self._logger.debug(message)

    def reset(self) -> None:
        """Resets the registry to an empty state."""
        self.root = {}

    def register_class(
        self,
        class_type: Type[Any],
        name: Optional[str] = None,
        should_warn_if_already_registered: bool = True,
    ) -> None:
        """Registers a class in the registry with a name."""
        key = name or class_type.__name__
        if key in self.root:
            if should_warn_if_already_registered:
                self.log(f"Class '{name}' already exists in registry")
        else:
            self.log(f"Registered new single class '{key}' in registry")
        self.root[key] = class_type

    def unregister_class(self, class_type: Type[Any]) -> None:
        """Unregisters a class from the registry."""
        key = class_type.__name__
        if key not in self.root:
            raise ClassRegistryNotFoundError(f"Class '{key}' not found in registry")
        del self.root[key]
        self.log(f"Unregistered single class '{key}' from registry")

    def unregister_class_by_name(self, name: str) -> None:
        """Unregisters a class from the registry by its name."""
        if name not in self.root:
            raise ClassRegistryNotFoundError(f"Class '{name}' not found in registry")
        del self.root[name]

    def register_classes_dict(self, classes: Dict[str, Type[Any]]) -> None:
        """Registers multiple classes in the registry with names."""
        self.root.update(classes)
        nb_classes = len(classes)
        if nb_classes > 1:
            self.log(f"Registered {len(classes)} classes in registry")
            classes_list_str = "\n".join([f"{key}: {value.__name__}" for key, value in classes.items()])
            logging.log(level=LOGGING_LEVEL_VERBOSE, msg=classes_list_str)
        else:
            self.log(f"Registered single class '{list(classes.values())[0].__name__}' in registry")

    def register_classes(self, classes: List[Type[Any]]) -> None:
        """Registers multiple classes in the registry with names."""
        if not classes:
            self.log("register_classes called with empty list of classes to register")
            return

        for class_type in classes:
            key = class_type.__name__
            if key in self.root:
                self.log(f"Class '{key}' already exists in registry, skipping")
                continue
            self.root[key] = class_type
        nb_classes = len(classes)
        if nb_classes > 1:
            self.log(f"Registered {nb_classes} classes in registry")
            classes_list_str = "\n".join([f"{the_class.__name__}: {the_class}" for the_class in classes])
            logging.log(level=LOGGING_LEVEL_VERBOSE, msg=classes_list_str)
        else:
            self.log(f"Registered single class '{classes[0].__name__}' in registry")

    def get_class(self, name: str) -> Optional[Type[Any]]:
        """Retrieves a class from the registry by its name. Returns None if not found."""
        return self.root.get(name)

    def get_required_class(self, name: str) -> Type[Any]:
        """Retrieves a class from the registry by its name. Raises an error if not found."""
        if name not in self.root:
            raise ClassRegistryNotFoundError(f"Class '{name}' not found in registry")
        return self.root[name]

    def get_required_subclass(self, name: str, base_class: Type[Any]) -> Type[Any]:
        """Retrieves a class from the registry by its name. Raises an error if not found."""
        if name not in self.root:
            raise ClassRegistryNotFoundError(f"Class '{name}' not found in registry")
        if not issubclass(self.root[name], base_class):
            raise ClassRegistryInheritanceError(f"Class '{name}' is not a subclass of {base_class}")
        return self.root[name]

    def get_required_base_model(self, name: str) -> Type[BaseModel]:
        return self.get_required_subclass(name=name, base_class=BaseModel)

    def has_class(self, name: str) -> bool:
        """Checks if a class is in the registry by its name."""
        return name in self.root

    def has_subclass(self, name: str, base_class: Type[Any]) -> bool:
        """Checks if a class is in the registry by its name."""
        if name not in self.root:
            return False
        if not issubclass(self.root[name], base_class):
            return False
        return True

    def register_classes_in_file(
        self,
        file_path: str,
        base_class: Optional[Type[Any]],
        is_include_imported: bool,
    ) -> None:
        """Processes a Python file to find and register classes."""
        module = import_module_from_file(file_path)

        # Find classes that match criteria
        classes_to_register = find_classes_in_module(
            module=module,
            base_class=base_class,
            include_imported=is_include_imported,
        )

        # Clean up sys.modules to prevent memory leaks
        del sys.modules[module.__name__]

        self.register_classes(classes=classes_to_register)

    def register_classes_in_folder(
        self,
        folder_path: str,
        base_class: Optional[Type[Any]] = None,
        is_recursive: bool = True,
        is_include_imported: bool = False,
    ) -> None:
        """
        Registers all classes in Python files within folders that are subclasses of base_class.
        If base_class is None, registers all classes.

        Args:
            folder_paths: List of paths to folders containing Python files
            base_class: Optional base class to filter registerable classes
            recursive: Whether to search recursively in subdirectories
            exclude_files: List of filenames to exclude
            exclude_dirs: List of directory names to exclude
            include_imported: Whether to include classes imported from other modules
        """

        python_files = find_files_in_dir(
            dir_path=folder_path,
            pattern="*.py",
            is_recursive=is_recursive,
        )

        for python_file in python_files:
            self.register_classes_in_file(
                file_path=str(python_file),
                base_class=base_class,
                is_include_imported=is_include_imported,
            )


class_registry = ClassRegistry()



# SPDX-FileCopyrightText: © 2018 Bastien Pietropaoli
# SPDX-FileCopyrightText: © 2025 Evotis S.A.S.
# SPDX-License-Identifier: Apache-2.0

"""
Copyright (c) 2018 Bastien Pietropaoli

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

All additions and modifications are Copyright (c) 2025 Evotis S.A.S.
"""

import importlib
import json
import logging
import types
import warnings
from typing import Any, Callable, ClassVar, Dict, Type, TypeVar

from pydantic import BaseModel, RootModel, ValidationError

from kajson.class_registry import class_registry
from kajson.exceptions import KajsonDecoderError
from kajson.sandbox_manager import sandbox_manager

DECODER_LOGGER_CHANNEL_NAME = "kajson.decoder"
DECODER_LOGGER_CHANNEL_NAME_IN_SANDBOX = "kajson.decoder.sandbox"
IS_DECODER_FALLBACK_ENABLED = False
FALLBACK_MESSAGE = " Trying something else."


T = TypeVar("T")


class UniversalJSONDecoder(json.JSONDecoder):
    """
    A universal JSON decoder for Python objects. To be used with JSON strings created
    using the `UniversalJSONEncoder`. Will use static methods `__json_decode__()` if
    provided in your custom classes. It will use the default JSON decoder whenever
    possible and then try multiple techniques to decode the provided objects.

    In addition, it is possible to register functions for types over which you have
    no control (standard / external library types). For this, use static method
    `UniversalJSONDecoder.register()`.

    How to use this class:
        `json.loads(s, cls=UniversalJSONDecoder)`
                        OR
        `kajson.loads(s)`
    """

    # The registered decoding functions:
    _decoders: ClassVar[Dict[Type[Any], Callable[[Dict[str, Any]], Any]]] = {}

    @staticmethod
    def register(obj_type: Type[T], decoding_function: Callable[[Dict[str, Any]], T]) -> None:
        """
        Register a function as a decoder for the provided type/class. The provided
        decoder should take a single argument (the raw dict to deserialise) and return
        an instance of that object.
        Passing a new decoding function to a type already registered will overwrite
        the previously registered decoding function.
        Args:
            type (obj_type): The type to be decoded by the provided decoder. Can be
                easily obtained by simply providing a class directly.
            decoding_function (function): The function to use as a decoder for the
                provided type. Takes a single argument, a returns an object.
        """
        if not isinstance(obj_type, type):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise ValueError("Expected a type/class, a %s was passed instead." % type(obj_type))
        if not callable(decoding_function):
            raise ValueError("Expected a function, a %s was passed instead." % type(decoding_function))

        UniversalJSONDecoder._decoders[obj_type] = decoding_function

    # Required to redirect the hook for decoding.
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Constructor redirecting the hook for decoding JSON objects."""
        json.JSONDecoder.__init__(self, object_hook=self.universal_decoder, *args, **kwargs)
        self.logger = logging.getLogger(DECODER_LOGGER_CHANNEL_NAME)

    def log(self, message: str) -> None:
        if sandbox_manager.is_in_sandbox():
            logger = logging.getLogger(DECODER_LOGGER_CHANNEL_NAME_IN_SANDBOX)
            logger.debug(message)
        else:
            self.logger.debug(message)

    def universal_decoder(self, the_dict: Dict[str, Any]) -> Any:
        """
        Universal decoder for JSON objects encoded using the UniversalJSONEncoder.
        It will try the following methods in order:
         - Default JSON encoder (for known types)
         - Registered encoding function (if one is found)
         - `__json_encode__()` as provided by the custom class (if it's found)
         - Use a constructor taking as argument the raw dictionary
         - Use the default constructor and replace the __dict__ property of the
           object (for custom classes)
        Args:
            the_dict (dict): A raw dictionnary obtained from the JSON string to be made
                into a beautiful Python object.
        Return:
            object - A Python object corresponding to the provided JSON object. If
                nothing could be done for that object, the raw dictionary is returned
                as is.
        """
        # Base object:
        if "__class__" not in the_dict:
            return the_dict

        # Get the class and module of the object:
        class_name = the_dict.pop("__class__")
        module_name = the_dict.pop("__module__")

        the_class: Type[Any]

        all_mods = _get_imported_modules()
        if module_name in all_mods:
            the_class = getattr(all_mods[module_name], class_name)
        elif registered_class := class_registry.get_class(name=class_name):
            self.log(f"Found class '{class_name}' in registry")
            the_class = registered_class
        else:
            self.log(f"Module '{module_name}' not found either in imported modules or registry, now importing")
            try:
                m = importlib.import_module(module_name)
            except Exception as exc:
                error_msg = f"Error while trying to import module {module_name}: {exc}"
                self.log(error_msg)
                raise KajsonDecoderError(error_msg) from exc
            self.log(f"Module '{module_name}' imported")
            the_class = getattr(m, class_name)

        # Registered decoder if any:
        if the_class in UniversalJSONDecoder._decoders:
            try:
                return UniversalJSONDecoder._decoders[the_class](the_dict)
            except Exception as exc:
                func_name = UniversalJSONDecoder._decoders[the_class].__name__
                error_msg = f"Decoding function '{func_name}' used for type '{the_class}' raised an exception: '{exc}'."
                if IS_DECODER_FALLBACK_ENABLED:
                    warnings.warn(error_msg + FALLBACK_MESSAGE)
                else:
                    raise KajsonDecoderError(error_msg) from exc

        # __json_decode__() static method if any:
        try:
            return getattr(the_class, "__json_decode__")(the_dict)
        except AttributeError:
            pass
        except Exception as exc:
            error_msg = f"Static method __json_deconde__ used for type '{the_class}' raised an exception: '{exc}'."
            if IS_DECODER_FALLBACK_ENABLED:
                warnings.warn(error_msg + FALLBACK_MESSAGE)
            else:
                raise KajsonDecoderError(error_msg) from exc

        if issubclass(the_class, RootModel):
            self.log(f"Using dictionary to initialize root of root_model '{the_class}'")
            # for soem reason, calling model_validate directly on the dict does not work:
            # as it builds another layer around the root: {"root": {"root": {}}
            try:
                self.log(f"Creating root model '{the_class}'...")
                root_model_obj: RootModel[Any] = the_class(**the_dict)
                self.log(f"Root model '{the_class}' created: {root_model_obj}")
            except ValidationError as exc:
                error_msg = f"Error: creating root_model '{the_class}': {exc}\n\nthe_dict:\n{the_dict}"
                self.log(error_msg)
                raise KajsonDecoderError(error_msg) from exc

            try:
                self.log(f"Trying to validate root model '{the_class}' with object '{root_model_obj}'...")
                the_class.model_validate(obj=root_model_obj)
                self.log(f"Root model '{the_class}' validated")
                return root_model_obj
            except ValidationError as exc:
                error_msg = f"Error: post validate root_model '{the_class}': {exc}\n\nthe_dict:\n{the_dict}\n\nroot_model_obj:\n{root_model_obj}"
                self.log(error_msg)
                raise KajsonDecoderError(error_msg) from exc

        if issubclass(the_class, BaseModel):
            self.log(f"Using model_validate for class '{the_class}'")
            try:
                return the_class.model_validate(the_dict)
            except ValidationError as exc:
                error_msg = f"Error: model_validate for '{the_class}': {exc}\n\nthe_dict:\n{the_dict}"
                self.log(error_msg)
                try:
                    base_model_obj = the_class(**the_dict)
                except ValidationError as exc:
                    error_msg = f"Error while trying to instantiate using kwargs '{the_class}': {exc}\n\nthe_dict:\n{the_dict}"
                    self.log(error_msg)
                    raise KajsonDecoderError(error_msg) from exc
                try:
                    self.log(f"Trying to validate base model '{the_class}' with object '{base_model_obj}'")
                    the_class.model_validate(obj=base_model_obj)
                    self.log(f"Base model '{the_class}' validated")
                    return base_model_obj
                except ValidationError as exc:
                    error_msg = (
                        f"Error while trying to post validate base_model '{the_class}': "
                        f"{exc}\n\nthe_dict:\n{the_dict}\n\nbase_model_obj:\n{base_model_obj}"
                    )
                    self.log(error_msg)
                    raise KajsonDecoderError(error_msg) from exc

        # Try the constructor with the dictionary as arguments:
        try:
            return the_class(**the_dict)
        except Exception:
            pass

        # Try the default constructor (no arguments)
        # and replace __dict__
        try:
            obj = the_class()
            obj.__dict__ = the_dict
            return obj
        except Exception:
            pass

        # Default, return the raw dict:
        return the_dict


#########################################################################################
#########################################################################################
#########################################################################################


def _get_imported_modules() -> Dict[str, types.ModuleType]:
    """
    Get all the already imported modules.
    Found here: http://stackoverflow.com/a/4858123/5321016
    Return:
        dict[str:module] - A dictionary of the already imported modules.
            Keys are modules' names.
    """
    result: Dict[str, types.ModuleType] = {}
    for attribute in globals().values():
        if isinstance(attribute, types.ModuleType):
            result[attribute.__name__] = attribute
    return result



# SPDX-FileCopyrightText: © 2018 Bastien Pietropaoli
# SPDX-FileCopyrightText: © 2025 Evotis S.A.S.
# SPDX-License-Identifier: Apache-2.0

"""
Copyright (c) 2018 Bastien Pietropaoli

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

All additions and modifications are Copyright (c) 2025 Evotis S.A.S.
"""

import datetime
import json
from typing import IO, Any, Dict, Union
from zoneinfo import ZoneInfo

from kajson.json_decoder import UniversalJSONDecoder
from kajson.json_encoder import UniversalJSONEncoder

# ------------------------------------------------
# API similar to the standard library json package
# ------------------------------------------------


def dumps(obj: Any, **kwargs: Any) -> str:
    """
    Serialise a given object into a JSON formatted string. This function
    uses the `UniversalJSONEncoder` instead of the default JSON encoder
    provided in the standard library. Takes the same keyword arguments as
    `json.dumps()` except for `cls` that is used to pass our custom encoder.
    Args:
        obj (object): The object to serialise.
        kwargs (**): Keyword arguments normally passed to `json.dumps()` except
            for `cls`. Unpredictable behaviour might occur if `cls` is passed.
    Return:
        str - The object serialised into a JSON string.
    """
    return json.dumps(obj, cls=UniversalJSONEncoder, **kwargs)


def dump(obj: Any, fp: IO[str], **kwargs: Any) -> None:
    """
    Serialise a given object into a JSON formatted file / stream. This function
    uses the `UniversalJSONEncoder` instead of the default JSON encoder provided
    in the standard library. Takes the same keyword arguments as `json.dump()`
    except for `cls` that is used to pass our custom encoder.
    Args:
        obj (object): The object to serialise.
        fp (file-like object): A .write()-supporting file-like object.
        kwargs (**): Keyword arguments normally passed to `json.dump()` except
            for `cls`. Unpredictable behaviour might occur if `cls` is passed.
    """
    json.dump(obj, fp, cls=UniversalJSONEncoder, **kwargs)


def loads(json_string: Union[str, bytes], **kwargs: Any) -> Any:
    """
    Deserialise a given JSON formatted str into a Python object using the
    `UniversalJSONDecoder`. Takes the same keyword arguments as `json.loads()`
    except for `cls` that is used to pass our custom decoder.
    Args:
        json_string (str): The JSON formatted string to decode.
        kwargs (**): Keyword arguments normally passed to `json.loads()` except
            for `cls`. Unpredictable behaviour might occur if `cls` is passed.
    Return:
        object - A Python object corresponding to the provided JSON formatted string.
    """
    return json.loads(json_string, cls=UniversalJSONDecoder, **kwargs)


def load(fp: IO[str], **kwargs: Any) -> Any:
    """
    Deserialise a given JSON formatted stream / file into a Python object using
    the `UniversalJSONDecoder`. Takes the same keyword arguments as `json.load()`
    except for `cls` that is used to pass our custom decoder.
    Args:
        fp (file-like object): A .write()-supporting file-like object.
        kwargs (**): Keyword arguments normally passed to `json.load()` except
            for `cls`. Unpredictable behaviour might occur if `cls` is passed.
    Return:
        object - A Python object corresponding to the provided JSON formatted stream / file.
    """
    return json.load(fp, cls=UniversalJSONDecoder, **kwargs)


#########################################################################################
#########################################################################################
#########################################################################################


# --------------------------------
# Some useful encoders / decoders:
# --------------------------------


# other implementation using more recent zoneinfo, without the need for pytz (untested):
def json_encode_timezone(t: ZoneInfo) -> Dict[str, Any]:
    """Encoder for timezones (using zoneinfo from Python 3.9+)."""
    return {"zone": t.key, "__class__": "timezone", "__module__": "zoneinfo"}


#########################################################################################
def json_encode_date(d: datetime.date) -> Dict[str, str]:
    """Encoder for dates (from module datetime)."""
    return {"date": str(d)}


UniversalJSONEncoder.register(datetime.date, json_encode_date)


def json_decode_date(d: Dict[str, str]) -> datetime.date:
    """Decoder for dates (from module datetime)."""
    # Split date string into parts and convert to integers
    year, month, day = map(int, d["date"].split("-"))
    return datetime.date(year, month, day)


UniversalJSONDecoder.register(datetime.date, json_decode_date)

#########################################################################################


def json_encode_datetime(d: datetime.datetime) -> Dict[str, Any]:
    """Encoder for datetimes (from module datetime)."""
    tzinfo = str(d.tzinfo) if d.tzinfo else None
    return {"datetime": d.strftime("%Y-%m-%d %H:%M:%S.%f"), "tzinfo": tzinfo, "__class__": "datetime", "__module__": "datetime"}


UniversalJSONEncoder.register(datetime.datetime, json_encode_datetime)


def json_decode_datetime(d: Dict[str, Any]) -> datetime.datetime:
    """Decoder for datetimes (from module datetime)."""
    dt = datetime.datetime.strptime(d["datetime"], "%Y-%m-%d %H:%M:%S.%f")
    if d.get("tzinfo"):
        try:
            dt = dt.replace(tzinfo=ZoneInfo(d["tzinfo"]))
        except Exception:
            # If timezone conversion fails, return naive datetime
            pass
    return dt


UniversalJSONDecoder.register(datetime.datetime, json_decode_datetime)

#########################################################################################


def json_encode_time(t: datetime.time) -> Dict[str, Any]:
    """Encoder for times (from module datetime)."""
    return {"time": t.strftime("%H:%M:%S.%f"), "tzinfo": t.tzinfo}


UniversalJSONEncoder.register(datetime.time, json_encode_time)


def json_decode_time(d: Dict[str, Any]) -> datetime.time:
    """Decoder for times (from module datetime)."""
    # Split time string into parts
    time_parts = d["time"].split(":")
    hours = int(time_parts[0])
    minutes = int(time_parts[1])
    # Handle seconds and milliseconds
    seconds_parts = time_parts[2].split(".")
    seconds = int(seconds_parts[0])
    milliseconds = int(seconds_parts[1])

    return datetime.time(hours, minutes, seconds, milliseconds, tzinfo=d["tzinfo"])


UniversalJSONDecoder.register(datetime.time, json_decode_time)

#########################################################################################


def json_encode_timedelta(t: datetime.timedelta) -> Dict[str, float]:
    """Encoder for timedeltas (from module datetime)."""
    return {"seconds": t.total_seconds()}


UniversalJSONEncoder.register(datetime.timedelta, json_encode_timedelta)
# Won't require a decoder since "seconds" will be automatically passed to a constructor.



# SPDX-FileCopyrightText: © 2025 Evotis S.A.S.
# SPDX-License-Identifier: Apache-2.0


class RootException(Exception):
    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class KajsonException(RootException):
    pass


class ClassRegistryInheritanceError(KajsonException):
    pass


class ClassRegistryNotFoundError(KajsonException):
    pass


class KajsonDecoderError(KajsonException):
    pass


class UnijsonEncoderError(KajsonException):
    pass



# SPDX-FileCopyrightText: © 2025 Evotis S.A.S.
# SPDX-License-Identifier: Apache-2.0

import importlib.util
import inspect
import os
import sys
from typing import Any, List, Optional, Type


class ModuleFileError(Exception):
    """Exception raised for errors related to module file operations."""

    pass


def import_module_from_file(file_path: str) -> Any:
    """Imports a module from a file path.

    Args:
        file_path: Path to the Python file to import

    Returns:
        The imported module

    Raises:
        ModuleFileError: If the file is not a Python file or cannot be loaded
    """
    # Validate that the file is a Python file
    if not file_path.endswith(".py"):
        raise ModuleFileError(f"File {file_path} is not a Python file (must end with .py)")

    module_name = file_path[:-3].replace(os.sep, ".")
    module = importlib.import_module(module_name)

    # Add the module to sys.modules to ensure proper imports within the module
    sys.modules[module_name] = module
    return module


def find_classes_in_module(
    module: Any,
    base_class: Optional[Type[Any]],
    include_imported: bool,
) -> List[Type[Any]]:
    """
    Finds all classes in a module that match the criteria.

    Args:
        module: The module to search for classes
        base_class: Optional base class to filter classes: will only return classes that are subclasses of this base_class
        include_imported: Whether to include classes imported from other modules

    Returns:
        List of class types that match the criteria
    """
    classes: List[Type[Any]] = []
    module_name = module.__name__

    # Find all classes in the module
    for _, obj in inspect.getmembers(module, inspect.isclass):
        # Skip classes that are imported from other modules
        if not include_imported and obj.__module__ != module_name:
            continue

        # Add the class if it's a subclass of base_class or if base_class is None
        if base_class is None or issubclass(obj, base_class):
            classes.append(obj)

    return classes



# SPDX-FileCopyrightText: © 2025 Evotis S.A.S.
# SPDX-License-Identifier: Apache-2.0

from typing import Callable, Optional

from pydantic import BaseModel


class SandboxManager(BaseModel):
    is_in_sandbox_callable: Optional[Callable[[], bool]] = None

    def set_sandbox_callable(self, sandbox_callable: Callable[[], bool]):
        """
        Sets a callable function to determine if we're in a sandbox, which would prevent importing unsafe modules,
        and in particular prevent logging with default methods.

        Args:
            safety_callable (Callable[[], bool]): A function that returns a boolean indicating if it's safe to log using default methods.
        """
        self.is_in_sandbox_callable = sandbox_callable

    def is_in_sandbox(self) -> bool:
        """
        Returns:
            bool: True if we're in a sandbox, False otherwise.
        """
        if self.is_in_sandbox_callable is None:
            # by default, we are not in a sandbox
            return False
        else:
            return self.is_in_sandbox_callable()


sandbox_manager = SandboxManager()
